unit CoreExports;

interface

uses
  Windows;

function AED_Connect(Host: PChar; bReadPort: LongBool; Index: LongInt): LongWord; cdecl;
function AED_SetFreq(uClientRef: LongWord; Hz: LongInt): LongWord; cdecl;
function AED_SetPacketBits(uClientRef: LongWord; NewPacketBits: LongInt): LongWord; cdecl;
function AED_SetBlockSize(uClientRef: LongWord; SizeBytes: LongInt): LongWord; cdecl;
function AED_SetAZero(uClientRef: LongWord; bNewBit: LongBool): LongWord; cdecl;
function AED_SetMDIO1(uClientRef, Index: LongWord; bNewBit: LongBool): LongWord; cdecl;
function AED_Go(uClientRef: LongWord): LongWord; cdecl;
function AED_GoClear(uClientRef: LongWord): LongWord; cdecl;
function AED_Clear(uClientRef: LongWord): LongWord; cdecl;
function AED_SetTimeout(uClientRef: LongWord; TimeoutMS: LongInt): LongWord; cdecl;
function AED_GetAllPackets(uClientRef: LongWord; pTimestampDataBytes: PLongInt; pTimestampData: PDouble; pPacketDataBytes: PLongInt; pPacketData: PByte): LongWord; cdecl;
function AED_GetLatestPacketAndDiscard(uClientRef: LongWord; pTimestampDataBytes: PLongInt; pTimestampData: PDouble; pPacketDataBytes: PLongInt; pPacketData: PByte): LongWord; cdecl;
function AED_GetData(uClientRef: LongWord; pDataBytes: PLongInt; pData: PByte): LongWord; cdecl;
function AED_PutData(uClientRef: LongWord; pDataBytes: PLongInt; pData: PByte): LongWord; cdecl;
function AED_GetStatus(uClientRef: LongWord): LongWord; cdecl;
function AED_Close(uClientRef: LongWord): LongWord; cdecl;





implementation

uses
  SysUtils, Classes, DIFFChunks, DClientSocks;

function IntAsStr(I: LongWord): String;
begin
  SetString(Result, PChar(@I), 4);
end;

function StrAsInt(S: String): LongWord;
begin
  case Length(S) of
    0: Result := 0;
    1: Result := PByte(@S[1])^;
    2: Result := PWord(@S[1])^;
    else Result := PLongWord(@S[1])^; //3 works because non-empty strings are null-terminated
  end;
end;

type
  TVClientSock = class
  protected
    hAPIMutex: THandle;
    ProCheck: LongWord;
  public
    ClientSock: TDClientSock;
    bRead: Boolean;
    InBuf2: String;
    hGotDataSem: THandle;
    Timeout: LongWord;
    PubCheck: LongWord;

    constructor Create;
    destructor Destroy; override;
    procedure ClientSockGotData(Sender: TObject);
    procedure TryRetry;
  end;

const
  ProData = $FEC72512;
  PubData = $93B36804;

{ TVClientSock }

procedure TVClientSock.ClientSockGotData(Sender: TObject);
begin
  ReleaseSemaphore(hGotDataSem, 1, nil);
end;

constructor TVClientSock.Create;
begin
  inherited;

  ClientSock := TDClientSock.Create;
  hGotDataSem := CreateSemaphore(nil, 0, 1, nil);
  hAPIMutex := CreateMutex(nil, False, nil);
end;

destructor TVClientSock.Destroy;
begin
  ClientSock.Free;
  CloseHandle(hGotDataSem);
  CloseHandle(hAPIMutex);

  inherited;
end;

procedure TVClientSock.TryRetry;
var
  pChunk: TDIFFChunk;
  ChunkBuf: String;
  I: Integer;

begin
  InBuf2 := '';

  pChunk := TDIFFChunk.Create;
  pChunk.TagStr := 'RETRY';
  ChunkBuf := pChunk.SaveToString;
  pChunk.Free;

  for I := 0 to 3 do begin
    Sleep(0);
    ClientSock.ReceiveStr; //Try to empty rest of garbage.
  end;
  ClientSock.PostSendStr(ChunkBuf);
end;

{ NotInAClass }
{ Compiler Doesn't Care }



function AED_Connect(Host: PChar; bReadPort: LongBool; Index: LongInt): LongWord; cdecl;
var
  TarHost: String;
  TarPort: Word;
  ThisClient: TVClientSock;
  Status: LongWord;

begin
  Result := 0;
  try
    TarHost := Host;
  except
    SetLastError(ERROR_NOACCESS);
    Exit;
  end;

  TarPort := $ED00 or (Index and $F);
  if bReadPort then TarPort := TarPort or $10;

  ThisClient := TVClientSock.Create;
  ThisClient.ProCheck := ProData;
  ThisClient.PubCheck := PubData;
  ThisClient.bRead := bReadPort;
  ThisClient.Timeout := 2000;
  ThisClient.ClientSock.OnGotData := ThisClient.ClientSockGotData;
  Status := ThisClient.ClientSock.ConnectTo(TarHost, TarPort);
  if Status <> ERROR_SUCCESS then begin
    ThisClient.Free;
    SetLastError(Status);
    Exit;
  end;

  Result := LongWord(Pointer(ThisClient));
  SetLastError(ERROR_SUCCESS);
end;

function VerifyClientRef(uClientRef: LongWord): LongWord;
begin
  try
    if uClientRef = 0 then
      Result := ERROR_BAD_UNIT
    else with TVClientSock(Pointer(uClientRef)) do
      if (ProCheck <> ProData)
      or (PubCheck <> PubData)
      then
        Result := ERROR_BAD_UNIT
      else
        Result := ERROR_SUCCESS
    ;
  except
    Result := ERROR_BAD_UNIT
  end;
end;

function AED_SetFreq(uClientRef: LongWord; Hz: LongInt): LongWord; cdecl;
const
  MaxHz = 40000000;

var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    if Hz < 0 then
      Hz := 0
    else if Hz > MaxHz then
      Hz := MaxHz
    ;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'WFREQ';
      pChunk.AddHeader('Hz', IntAsStr(Hz));
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    ThisClient.ClientSock.PostSendStr(DataBuf);

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_SetPacketBits(uClientRef: LongWord; NewPacketBits: LongInt): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    if NewPacketBits <= 0 then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'WBITS';
      pChunk.AddHeader('Bits', IntAsStr(NewPacketBits));
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    ThisClient.ClientSock.PostSendStr(DataBuf);

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_SetBlockSize(uClientRef: LongWord; SizeBytes: LongInt): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    if SizeBytes <= 0 then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'WBKSZ';
      pChunk.AddHeader('Bytes', IntAsStr(SizeBytes));
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    ThisClient.ClientSock.PostSendStr(DataBuf);

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_SetAZero(uClientRef: LongWord; bNewBit: LongBool): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'WDIO1';
      if bNewBit then
        pChunk.AddHeader(#0, #1)
      else
        pChunk.AddHeader(#0, #0)
      ;
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    ThisClient.ClientSock.PostSendStr(DataBuf);

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_SetMDIO1(uClientRef, Index: LongWord; bNewBit: LongBool): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    if Index >= 32 then begin
      Result := ERROR_INVALID_INDEX;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'WDIO1';
      if bNewBit then
        pChunk.AddHeader(Char(Index), #1)
      else
        pChunk.AddHeader(Char(Index), #0)
      ;
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    ThisClient.ClientSock.PostSendStr(DataBuf);

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_Go(uClientRef: LongWord): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'GOGO.';
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    ThisClient.ClientSock.PostSendStr(DataBuf);

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_GoClear(uClientRef: LongWord): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'GOCLR';
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    ThisClient.ClientSock.PostSendStr(DataBuf);

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_Clear(uClientRef: LongWord): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'CLEAR';
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    ThisClient.ClientSock.PostSendStr(DataBuf);

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_SetTimeout(uClientRef: LongWord; TimeoutMS: LongInt): LongWord; cdecl;
var
  ThisClient: TVClientSock;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      Exit;
    end;

    if TimeoutMS <= 0 then begin
      Result := ERROR_BAD_LENGTH;
      Exit;
    end;

    ThisClient.Timeout := TimeoutMS;

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_GetAllPackets(uClientRef: LongWord; pTimestampDataBytes: PLongInt; pTimestampData: PDouble; pPacketDataBytes: PLongInt; pPacketData: PByte): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;
  bTriedRetry: Boolean;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    pTimestampDataBytes^ := 0;
    pPacketDataBytes^ := 0;
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.bRead then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      pTimestampDataBytes^ := 0;
      pPacketDataBytes^ := 0;
      Exit;
    end;

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      pTimestampDataBytes^ := 0;
      pPacketDataBytes^ := 0;
      Exit;
    end;

    if (pTimestampDataBytes^ <= 0) or (pPacketDataBytes^ <= 0) then begin
      Result := ERROR_BAD_LENGTH;
      pTimestampDataBytes^ := 0;
      pPacketDataBytes^ := 0;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'RREQA';
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    while WaitForSingleObject(ThisClient.hGotDataSem, 0) = WAIT_OBJECT_0 do ;
    ThisClient.InBuf2 := ThisClient.InBuf2 + ThisClient.ClientSock.ReceiveStr;
    ThisClient.ClientSock.PostSendStr(DataBuf);

    bTriedRetry := False;
    Result := ERROR_TIMEOUT;
    repeat
      if Length(ThisClient.InBuf2) = 0 then
        pChunk := nil
      else try
        pChunk := ExtractChunk(ThisClient.InBuf2);
      except
        Result := ERROR_READ_FAULT;
        Break;
      end;
      if pChunk = nil then begin
        if WaitForSingleObject(ThisClient.hGotDataSem, ThisClient.Timeout) = WAIT_TIMEOUT then Break;
        ThisClient.InBuf2 := ThisClient.InBuf2 + ThisClient.ClientSock.ReceiveStr;
        try
          pChunk := ExtractChunk(ThisClient.InBuf2);
        except
          Result := ERROR_READ_FAULT;
          Break;
        end;
      end;
      if pChunk <> nil then try
        if pChunk.TagStr = 'RDATA' then begin
          Result := ERROR_SUCCESS;

          if Length(pChunk.SpecData) < pPacketDataBytes^ then pPacketDataBytes^ := Length(pChunk.SpecData);
          Move(pChunk.SpecData[1], pPacketData^, pPacketDataBytes^);

          DataBuf := pChunk.HeaderByName['Time'];
          if Length(DataBuf) < pTimestampDataBytes^ then pTimestampDataBytes^ := Length(DataBuf);
          Move(DataBuf[1], pTimestampData^, pTimestampDataBytes^);

          Break;
        end
        else if pChunk.TagStr = 'NOWRK' then begin
          Result := ERROR_BAD_NET_RESP;
          pTimestampDataBytes^ := 0;
          pPacketDataBytes^ := 0;
          SetLastError(StrAsInt(pChunk.HeaderByName['Status']));
          Break;
        end
        else if pChunk.TagStr = 'IOERR' then begin
          Result := ERROR_BAD_DEV_TYPE;
          pTimestampDataBytes^ := 0;
          pPacketDataBytes^ := 0;
          Break;
        end
        else if not bTriedRetry then begin
          bTriedRetry := True;
          ThisClient.TryRetry;
        end
        else begin
          Result := ERROR_INVALID_DATA;
          pTimestampDataBytes^ := 0;
          pPacketDataBytes^ := 0;
          ThisClient.InBuf2 := '';
          Break;
        end;
      finally
        pChunk.Free;
      end;
    until False;
    if (Result = ERROR_TIMEOUT) and not ThisClient.ClientSock.bConnected then Result := ERROR_NOT_CONNECTED;

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_GetLatestPacketAndDiscard(uClientRef: LongWord; pTimestampDataBytes: PLongInt; pTimestampData: PDouble; pPacketDataBytes: PLongInt; pPacketData: PByte): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;
  bTriedRetry: Boolean;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    pTimestampDataBytes^ := 0;
    pPacketDataBytes^ := 0;
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.bRead then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      pTimestampDataBytes^ := 0;
      pPacketDataBytes^ := 0;
      Exit;
    end;

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      pTimestampDataBytes^ := 0;
      pPacketDataBytes^ := 0;
      Exit;
    end;

    if (pTimestampDataBytes^ <= 0) or (pPacketDataBytes^ <= 0) then begin
      Result := ERROR_BAD_LENGTH;
      pTimestampDataBytes^ := 0;
      pPacketDataBytes^ := 0;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'RREQL';
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    while WaitForSingleObject(ThisClient.hGotDataSem, 0) = WAIT_OBJECT_0 do ;
    ThisClient.InBuf2 := ThisClient.InBuf2 + ThisClient.ClientSock.ReceiveStr;
    ThisClient.ClientSock.PostSendStr(DataBuf);

    bTriedRetry := False;
    Result := ERROR_TIMEOUT;
    repeat
      if Length(ThisClient.InBuf2) = 0 then
        pChunk := nil
      else try
        pChunk := ExtractChunk(ThisClient.InBuf2);
      except
        Result := ERROR_READ_FAULT;
        Break;
      end;
      if pChunk = nil then begin
        if WaitForSingleObject(ThisClient.hGotDataSem, ThisClient.Timeout) = WAIT_TIMEOUT then Break;
        ThisClient.InBuf2 := ThisClient.InBuf2 + ThisClient.ClientSock.ReceiveStr;
        try
          pChunk := ExtractChunk(ThisClient.InBuf2);
        except
          Result := ERROR_READ_FAULT;
          Break;
        end;
      end;
      if pChunk <> nil then try
        if pChunk.TagStr = 'RDATA' then begin
          Result := ERROR_SUCCESS;

          if Length(pChunk.SpecData) < pPacketDataBytes^ then pPacketDataBytes^ := Length(pChunk.SpecData);
          Move(pChunk.SpecData[1], pPacketData^, pPacketDataBytes^);

          DataBuf := pChunk.HeaderByName['Time'];
          if Length(DataBuf) < pTimestampDataBytes^ then pTimestampDataBytes^ := Length(DataBuf);
          Move(DataBuf[1], pTimestampData^, pTimestampDataBytes^);

          Break;
        end
        else if pChunk.TagStr = 'NOWRK' then begin
          Result := ERROR_BAD_NET_RESP;
          pTimestampDataBytes^ := 0;
          pPacketDataBytes^ := 0;
          SetLastError(StrAsInt(pChunk.HeaderByName['Status']));
          Break;
        end
        else if pChunk.TagStr = 'IOERR' then begin
          Result := ERROR_BAD_DEV_TYPE;
          pTimestampDataBytes^ := 0;
          pPacketDataBytes^ := 0;
          Break;
        end
        else if not bTriedRetry then begin
          bTriedRetry := True;
          ThisClient.TryRetry;
        end
        else begin
          Result := ERROR_INVALID_DATA;
          pTimestampDataBytes^ := 0;
          pPacketDataBytes^ := 0;
          ThisClient.InBuf2 := '';
          Break;
        end;
      finally
        pChunk.Free;
      end;
    until False;
    if (Result = ERROR_TIMEOUT) and not ThisClient.ClientSock.bConnected then Result := ERROR_NOT_CONNECTED;

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_GetData(uClientRef: LongWord; pDataBytes: PLongInt; pData: PByte): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;
  bTriedRetry: Boolean;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    pDataBytes^ := 0;
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if not ThisClient.bRead then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      pDataBytes^ := 0;
      Exit;
    end;

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      pDataBytes^ := 0;
      Exit;
    end;

    if pDataBytes^ <= 0 then begin
      Result := ERROR_BAD_LENGTH;
      pDataBytes^ := 0;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'RREQ?';
      pChunk.AddHeader('Len', IntAsStr(pDataBytes^));
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      FreeAndNil(pChunk);
      Exit;
    end;

    while WaitForSingleObject(ThisClient.hGotDataSem, 0) = WAIT_OBJECT_0 do ;
    ThisClient.InBuf2 := ThisClient.InBuf2 + ThisClient.ClientSock.ReceiveStr;
    ThisClient.ClientSock.PostSendStr(DataBuf);

    bTriedRetry := False;
    Result := ERROR_TIMEOUT;
    repeat
      if Length(ThisClient.InBuf2) = 0 then
        pChunk := nil
      else try
        pChunk := ExtractChunk(ThisClient.InBuf2);
      except
        Result := ERROR_READ_FAULT;
        Break;
      end;
      if pChunk = nil then begin
        if WaitForSingleObject(ThisClient.hGotDataSem, ThisClient.Timeout) = WAIT_TIMEOUT then Break;
        ThisClient.InBuf2 := ThisClient.InBuf2 + ThisClient.ClientSock.ReceiveStr;
        try
          pChunk := ExtractChunk(ThisClient.InBuf2);
        except
          Result := ERROR_READ_FAULT;
          Break;
        end;
      end;
      if pChunk <> nil then try
        if pChunk.TagStr = 'RDATA' then begin
          Result := ERROR_SUCCESS;
          if Length(pChunk.SpecData) < pDataBytes^ then pDataBytes^ := Length(pChunk.SpecData);
          Move(pChunk.SpecData[1], pData^, pDataBytes^);
          Break;
        end
        else if pChunk.TagStr = 'NOWRK' then begin
          Result := ERROR_BAD_NET_RESP;
          pDataBytes^ := 0;
          SetLastError(StrAsInt(pChunk.HeaderByName['Status']));
          Break;
        end
        else if pChunk.TagStr = 'IOERR' then begin
          Result := ERROR_BAD_DEV_TYPE;
          pDataBytes^ := 0;
          Break;
        end
        {
        else if (pChunk.TagStr = #0#0#0#0#0) and (Length(pChunk.SpecData) = 0) then begin
        end
        else begin
          DumpLog(pChunk.SpecData, ' ' + StringReplaceC(pChunk.TagStr, #0, '') + '.bin');
        end;
        }
        else if not bTriedRetry then begin
          bTriedRetry := True;
          ThisClient.TryRetry;
        end
        else begin
          Result := ERROR_INVALID_DATA;
          pDataBytes^ := 0;
          ThisClient.InBuf2 := '';
          Break;
        end;
      finally
        pChunk.Free;
      end;
    until False;
    if (Result = ERROR_TIMEOUT) and not ThisClient.ClientSock.bConnected then Result := ERROR_NOT_CONNECTED;

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_PutData(uClientRef: LongWord; pDataBytes: PLongInt; pData: PByte): LongWord; cdecl;
var
  ThisClient: TVClientSock;
  DataBuf: String;
  pChunk: TDIFFChunk;
  bTriedRetry: Boolean;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then begin
    pDataBytes^ := 0;
    Exit;
  end;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  try

    if ThisClient.bRead then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      pDataBytes^ := 0;
      Exit;
    end;

    if not ThisClient.ClientSock.bConnected then begin
      Result := ERROR_NOT_CONNECTED;
      pDataBytes^ := 0;
      Exit;
    end;

    if pDataBytes^ <= 0 then begin
      Result := ERROR_BAD_LENGTH;
      pDataBytes^ := 0;
      Exit;
    end;

    try
      SetString(DataBuf, PChar(pData), pDataBytes^);
    except
      Result := ERROR_NOACCESS;
      pDataBytes^ := 0;
      Exit;
    end;

    pChunk := nil;
    try
      pChunk := TDIFFChunk.Create;
      pChunk.TagStr := 'WDATA';
      pChunk.SpecData := DataBuf;
      DataBuf := pChunk.SaveToString;
      FreeAndNil(pChunk);
    except
      Result := ERROR_INTERNAL_ERROR;
      pDataBytes^ := 0;
      FreeAndNil(pChunk);
      Exit;
    end;

    while WaitForSingleObject(ThisClient.hGotDataSem, 0) = WAIT_OBJECT_0 do ;
    ThisClient.InBuf2 := ThisClient.InBuf2 + ThisClient.ClientSock.ReceiveStr;
    ThisClient.ClientSock.PostSendStr(DataBuf);

    bTriedRetry := False;
    Result := ERROR_TIMEOUT;
    repeat
      if Length(ThisClient.InBuf2) = 0 then
        pChunk := nil
      else try
        pChunk := ExtractChunk(ThisClient.InBuf2);
      except
        Result := ERROR_READ_FAULT;
        Break;
      end;
      if pChunk = nil then begin
        if WaitForSingleObject(ThisClient.hGotDataSem, ThisClient.Timeout) = WAIT_TIMEOUT then Break;
        ThisClient.InBuf2 := ThisClient.InBuf2 + ThisClient.ClientSock.ReceiveStr;
        try
          pChunk := ExtractChunk(ThisClient.InBuf2);
        except
          Result := ERROR_READ_FAULT;
          Break;
        end;
      end;
      if pChunk <> nil then try
        if pChunk.TagStr = 'WRESP' then begin
          Result := ERROR_SUCCESS;
          pDataBytes^ := StrAsInt(pChunk.HeaderByName['Bytes']);
          Break;
        end
        else if pChunk.TagStr = 'NOWRK' then begin
          Result := ERROR_BAD_NET_RESP;
          pDataBytes^ := 0;
          SetLastError(StrAsInt(pChunk.HeaderByName['Status']));
          Break;
        end
        else if pChunk.TagStr = 'IOERR' then begin
          Result := ERROR_BAD_DEV_TYPE;
          pDataBytes^ := 0;
          Break;
        end
        else if not bTriedRetry then begin
          bTriedRetry := True;
          ThisClient.TryRetry;
        end
        else begin
          Result := ERROR_INVALID_DATA;
          pDataBytes^ := 0;
          ThisClient.InBuf2 := '';
          Break;
        end;
      finally
        pChunk.Free;
      end;
    until False;
    if (Result = ERROR_TIMEOUT) and not ThisClient.ClientSock.bConnected then Result := ERROR_NOT_CONNECTED;

  finally
    ReleaseMutex(ThisClient.hAPIMutex);
  end;
end;

function AED_GetStatus(uClientRef: LongWord): LongWord; cdecl;
begin
  Result := ERROR_NOT_SUPPORTED;
end;

function AED_Close(uClientRef: LongWord): LongWord; cdecl;
var
  ThisClient: TVClientSock;

begin
  Result := VerifyClientRef(uClientRef);
  if Result <> ERROR_SUCCESS then Exit;

  ThisClient := Pointer(uClientRef);
  WaitForSingleObject(ThisClient.hAPIMutex, INFINITE);
  ReleaseMutex(ThisClient.hAPIMutex);

  ThisClient.ProCheck := 0;
  ThisClient.PubCheck := 0;
  ThisClient.Free;
end;





end.
