unit DServerSocks;

interface

uses
  Windows, SysUtils, Classes, WinSock;

type
  TDServerSock = class;
  TDSSListenThread = class;
  TDSSServerClient = class;

  TDSSServerClients = array of TDSSServerClient;

  TDSSClientEvent = procedure (Sender: TObject; ServerClient: TDSSServerClient) of object;

  TDServerSock = class
  protected
    FListenThread: TDSSListenThread;
    FServerClient: TDSSServerClients;
    hClientMutex: THandle;
    function GetServerClients: TDSSServerClients; virtual;
  public
    ListenPort: Word;
    OnClientConnect, OnClientGotData, OnClientDisconnect: TDSSClientEvent;
    property ListenThread: TDSSListenThread read FListenThread;
    property ServerClients: TDSSServerClients read GetServerClients;
    constructor Create; virtual;
    destructor Destroy; override;
    procedure GoListen; virtual;
    function CleanupListen: Integer; virtual;
    procedure CleanupThreads; virtual;
  end;

  TDSSListenThread = class(TThread)
  protected
    hSocket: TSocket;
    FListenPort: Word;
    SyncEvent: TDSSClientEvent;
    procedure Execute; override;
  public
    MySock: TDServerSock;
  end;

  TDSSServerClientReadThread = class(TThread)
  protected
    MyServerClient: TDSSServerClient;
    procedure Execute; override;
  end;

  TDSSServerClientWriteThread = class(TThread)
  protected
    MyServerClient: TDSSServerClient;
    procedure Execute; override;
  end;

  TDSSServerClient = class
  protected
    hSocket: TSocket;
    SyncEvent: TDSSClientEvent;
    InBuf: String;
    hInBufMutex, hSendMutex, hSendSem: THandle;
    ReadThread: TDSSServerClientReadThread;
    WriteThread: TDSSServerClientWriteThread;
    PostedSend: array of String;
    
    procedure DoClientEvent; virtual;
  public
    MySock: TDServerSock;
    RemoteAddress: String;

    function SendStr(Buf: String): Integer;
    procedure PostSendStr(Buf: String);
    function GetPostedSend: String;
    //function TrySendStr(Buf: String): Integer;
    //function TryFlushPostSendStr: Boolean;
    function ReceiveStr: String;
    procedure DropConnection;
    procedure Resume;
    constructor Create; virtual;
    destructor Destroy; override;
  end;



implementation

const
  SD_BOTH = 2;

var
  WSAData: TWSAData;

function IntAsStr(I: LongWord): String;
begin
  SetString(Result, PChar(@I), 4);
end;

function MotorolaWord(W: Word): Word;
asm
  xchg AL, AH
end;

function RandHex: Char;
var
  N: Byte;
  
begin
  N := Random(16);
  if N < 10 then
    Result := Char(Byte('0') + N)
  else
    Result := Char(Byte('A') - 10 + N)
  ;
end;

{ TDServerSock }

function TDServerSock.CleanupListen: Integer;
begin
  Result := -1;
  if ListenThread = nil then Exit;
  if WaitForSingleObject(ListenThread.Handle, 0) <> WAIT_OBJECT_0 then Exit;
  GetExitCodeThread(ListenThread.Handle, LongWord(Result));
  if Result < 0 then Result := 0;
  ListenThread.Free;
  FListenThread := nil;
end;

procedure TDServerSock.CleanupThreads;
var
  H, I: Integer;

begin
  WaitForSingleObject(hClientMutex, INFINITE);
  try
    H := High(FServerClient);
    for I := H downto 0 do
      if WaitForSingleObject(FServerClient[I].ReadThread.Handle, 0) = WAIT_OBJECT_0 then begin
        FServerClient[I].Free;
        if I <> H then FServerClient[I] := FServerClient[H];
        SetLength(FServerClient, H);
        Dec(H);
      end
    ;
  finally
    ReleaseMutex(hClientMutex);
  end;
end;

constructor TDServerSock.Create;
begin
  inherited;

  hClientMutex := CreateMutex(nil, False, nil);
end;

destructor TDServerSock.Destroy;
var
  I: Integer;

begin
  if FListenThread <> nil then begin
    shutdown(FListenThread.hSocket, SD_BOTH);
    closesocket(FListenThread.hSocket);
  end;

  WaitForSingleObject(hClientMutex, INFINITE);
  try
    for I := 0 to High(FServerClient) do
      shutdown(FServerClient[I].hSocket, SD_BOTH)
    ;
  finally
    ReleaseMutex(hClientMutex);
  end;
  Sleep(0);
  WaitForSingleObject(hClientMutex, INFINITE);
  try
    for I := 0 to High(FServerClient) do
      FServerClient[I].Free
    ;
    SetLength(FServerClient, 0);
  finally
    ReleaseMutex(hClientMutex);
  end;

  CloseHandle(hClientMutex);

  FreeAndNil(FListenThread);

  inherited;
end;

function TDServerSock.GetServerClients: TDSSServerClients;
begin
  WaitForSingleObject(hClientMutex, INFINITE);
  try
    Result := FServerClient;
  finally
    ReleaseMutex(hClientMutex);
  end;
end;

procedure TDServerSock.GoListen;
begin
  FListenThread := TDSSListenThread.Create(True);
  FListenThread.FListenPort := ListenPort;
  FListenThread.MySock := Self;
  FListenThread.Resume;
end;

{ TDSSListenThread }

procedure TDSSListenThread.Execute;
var
  IPAddr: TSockAddrIn;
  L: Integer;
  hNewSock: TSocket;
  NewClient: TDSSServerClient;

  SockAddrIn: TSockAddrIn;
  SockAddrInSize: Integer;

begin
  hSocket := socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if hSocket = INVALID_SOCKET then begin
    ReturnValue := ERROR_NO_NET_OR_BAD_PATH;
    Exit;
  end;
  IPAddr.sin_family := AF_INET;
  IPAddr.sin_port := MotorolaWord(FListenPort);
  IPAddr.sin_addr.S_addr := INADDR_ANY;
  if bind(hSocket, IPAddr, SizeOf(IPAddr)) <> 0 then begin
    closesocket(hSocket);
    ReturnValue := ERROR_CONNECTION_UNAVAIL;
    Exit;
  end;
  if listen(hSocket, SOMAXCONN) <> 0 then begin
    closesocket(hSocket);
    ReturnValue := ERROR_NOT_READY;
    Exit;
  end;

  repeat
    L := SizeOf(IPAddr);
    hNewSock := accept(hSocket, @IPAddr, @L);
    if hNewSock = INVALID_SOCKET then Break;

    NewClient := TDSSServerClient.Create;
    NewClient.MySock := MySock;
    NewClient.hSocket := hNewSock;
    SockAddrInSize := SizeOf(SockAddrIn);
    if getpeername(hNewSock, SockAddrIn, SockAddrInSize) = 0 then
      NewClient.RemoteAddress := inet_ntoa(SockAddrIn.sin_addr)
    else
      NewClient.RemoteAddress := '?'
    ;

    NewClient.SyncEvent := MySock.OnClientConnect;
    if Assigned(NewClient.SyncEvent) then Synchronize(NewClient.DoClientEvent);

    NewClient.Resume;

    WaitForSingleObject(MySock.hClientMutex, INFINITE);
    try
      L := Length(MySock.FServerClient);
      SetLength(MySock.FServerClient, L + 1);
      MySock.FServerClient[L] := NewClient;
    finally
      ReleaseMutex(MySock.hClientMutex);
    end;
  until False;
  shutdown(hSocket, SD_BOTH);
  closesocket(hSocket);
  ReturnValue := ERROR_NOT_CONNECTED;
end;

{ TDSSServerClient }

constructor TDSSServerClient.Create;
begin
  inherited;

  InBuf := '';
  hInBufMutex := CreateMutex(nil, False, nil);
  hSendMutex := CreateMutex(nil, False, nil);
  hSendSem := CreateSemaphore(nil, 0, $0FFFFFFF, nil);
  ReadThread := TDSSServerClientReadThread.Create(True);
  ReadThread.MyServerClient := Self;
  WriteThread := TDSSServerClientWriteThread.Create(True);
  WriteThread.MyServerClient := Self;
end;

destructor TDSSServerClient.Destroy;
begin
  CloseHandle(hInBufMutex);
  CloseHandle(hSendMutex);
  CloseHandle(hSendSem);
  ReadThread.Free;
  WriteThread.Free;

  inherited;
end;

procedure TDSSServerClient.DoClientEvent;
begin
  try
    SyncEvent(MySock, Self);
  except
  end;
end;

procedure TDSSServerClient.DropConnection;
begin
  shutdown(hSocket, SD_BOTH);
end;

function TDSSServerClient.GetPostedSend: String;
var
  L: Integer;

begin
  WaitForSingleObject(hSendSem, INFINITE);

  WaitForSingleObject(hSendMutex, INFINITE);
  try
    L := Length(PostedSend);
    if L = 0 then begin
      Result := '';
    end
    else begin
      Result := PostedSend[0];
      Dec(L);
      PostedSend[0] := PostedSend[L]; //For the garbage collection.
      Move(PostedSend[1], PostedSend[0], L * 4);
      SetLength(PostedSend, L);
    end;
  finally
    ReleaseMutex(hSendMutex);
  end;
end;

procedure TDSSServerClient.PostSendStr(Buf: String);
var
  I: Integer;

begin
  if Length(Buf) = 0 then Exit; //We aren't datagrammy, don't bother with transport packets.

  WaitForSingleObject(hSendMutex, INFINITE);
  try
    I := Length(PostedSend);
    SetLength(PostedSend, I + 1);
    PostedSend[I] := Buf;
  finally
    ReleaseMutex(hSendMutex);
  end;

  ReleaseSemaphore(hSendSem, 1, nil);
end;

function TDSSServerClient.ReceiveStr: String;
begin
  WaitForSingleObject(hInBufMutex, INFINITE);
  try
    Result := InBuf;
    InBuf := '';
  finally
    ReleaseMutex(hInBufMutex);
  end;
end;

procedure TDSSServerClient.Resume;
begin
  ReadThread.Resume;
  WriteThread.Resume;
end;

function TDSSServerClient.SendStr(Buf: String): Integer;
begin
  try
    repeat
      Result := send(hSocket, Buf[1], Length(Buf), 0);
      if Result <= 0 then Break;
      if Result >= Length(Buf) then Break;
      Buf := Copy(Buf, Result + 1, MAXINT);
    until False;
    Buf := '';
  except
    Result := SOCKET_ERROR;
  end;
end;

{
function TDSSServerClient.TryFlushPostSendStr: Boolean;
var
  L, R: Integer;
  Buf: String;

begin
  L := Length(PostedSend);
  Result := L <> 0;
  if not Result then Exit;

  Buf := PostedSend[0];
  R := TrySendStr(Buf);
  if R < Length(Buf) then Exit;
  (*
  repeat
    R := TrySendStr(Buf);
    if R <= 0 then Exit;
    if R >= Length(Buf) then Break;
  until False;
  *)

  Dec(L);
  PostedSend[0] := PostedSend[L]; //For the garbage collection.
  Move(PostedSend[1], PostedSend[0], L * 4);
  SetLength(PostedSend, L);
end;

function TDSSServerClient.TrySendStr(Buf: String): Integer;
begin
  try
    Result := send(hSocket, Buf[1], Length(Buf), 0);
    Buf := '';
  except
    Result := SOCKET_ERROR;
  end;
end;
}

{ TDSSServerClientReadThread }

procedure TDSSServerClientReadThread.Execute;
const
  BufBlockSize = 65536;

var
  Buf: String;
  L: Integer;

begin
  with MyServerClient do begin
    try
      repeat
        SetString(Buf, nil, BufBlockSize);
        L := recv(hSocket, Buf[1], Length(Buf), 0);
        if L <= 0 then begin //Cope with violations of the spec.
          ReturnValue := ERROR_NOT_CONNECTED;
          Break;
        end;
        if L < BufBlockSize then SetLength(Buf, L);
        WaitForSingleObject(hInBufMutex, INFINITE);
        try
          InBuf := InBuf + Buf;
        finally
          ReleaseMutex(hInBufMutex);
        end;

        SyncEvent := MySock.OnClientGotData;
        if Assigned(SyncEvent) then Synchronize(DoClientEvent);
      until False;
    except
    end;

    ReleaseSemaphore(hSendSem, 1, nil);

    shutdown(hSocket, SD_BOTH);
    closesocket(hSocket);

    SyncEvent := MySock.OnClientDisconnect;
    if Assigned(SyncEvent) then Synchronize(DoClientEvent);
  end;
end;

{ TDSSServerClientWriteThread }

procedure TDSSServerClientWriteThread.Execute;
var
  Buf: String;
  I, L: Integer;

begin
  with MyServerClient do begin
    try
      repeat
        Buf := GetPostedSend;
        if Length(Buf) = 0 then begin
          ReturnValue := ERROR_CANCELLED;
          Break;
        end;

        I := 1;
        repeat
          L := Length(Buf) - I + 1;
          try
            L := send(hSocket, Buf[I], L, 0);
          except
            try
              L := send(hSocket, Buf[I], L, 0);
            except
              ReturnValue := ERROR_NOACCESS;
              Break;
            end;
          end;
          if L < 0 then L := 0; //Cope with violations of the spec.
          Inc(I, L);
        until (I >= Length(Buf)) or (L = 0);
        Buf := '';

        if L = 0 then begin
          ReturnValue := ERROR_NOT_CONNECTED;
          Break;
        end;
      until False;
    except
    end;

    shutdown(hSocket, SD_BOTH);
  end;
end;





initialization
  WSAStartup($101, WSAData);

finalization
  WSACleanup;
end.
