unit ServUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

const
  AM_PUMP = WM_USER + $99;

type
  TServForm = class(TForm)
    ViewTime: TTimer;
    ViewCheck: TCheckBox;
    ViewList: TListBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ViewCheckClick(Sender: TObject);
    procedure ViewTimeTimer(Sender: TObject);
  protected
    procedure IdleCheckPorts(Sender: TObject; var Done: Boolean);
    //procedure AMPump(var Msg: TMessage); message AM_PUMP;
  private
    { Private declarations }
  public
    procedure DoPump;
    { Public declarations }
  end;

var
  ServForm: TServForm;

implementation

{$R *.DFM}

uses
  PortServers;

const
  iEID = $00A5ED10;
  oEID = $00A5ED00;

var
  PortServ: array[0..3] of TPortServer;

procedure TServForm.DoPump;
begin
  PostMessage(Handle, AM_PUMP, $5A535CB8, $5F2AAEC1);
end;

procedure TServForm.FormCreate(Sender: TObject);
var
  I: Integer;

begin
  ViewCheck.Align := alTop;

  for I := 0 to 2 do begin
    PortServ[I] := TReadPortServer.Create;
    PortServ[I].Init(iEID or Byte(I), True);
  end;
  PortServ[3] := TWritePortServer.Create;
  PortServ[3].Init(oEID, False);

  Application.OnIdle := IdleCheckPorts;
end;

procedure TServForm.FormDestroy(Sender: TObject);
var
  I: Integer;
  
begin
  for I := 0 to High(PortServ) do PortServ[I].Free;
end;

procedure TServForm.IdleCheckPorts(Sender: TObject; var Done: Boolean);
var
  I: Integer;

begin
  if bNeedsCleanup then begin
    bNeedsCleanup := False;

    for I := 0 to High(PortServ) do
      PortServ[I].DoCleanup
    ;
  end;

  for I := 0 to High(PortServ) do begin
    PortServ[I].CheckData;
    if PortServ[I].bGotData then Done := False;
  end;
end;

procedure TServForm.ViewCheckClick(Sender: TObject);
begin
  if ViewCheck.Checked then begin
    ViewTime.Enabled := True;
    ViewList.Items.Clear;
    ViewList.Visible := True;
    ViewTime.OnTimer(ViewTime);
  end
  else begin
    ViewTime.Enabled := False;
    ViewList.Visible := False;
  end;
end;

procedure TServForm.ViewTimeTimer(Sender: TObject);
const
  bReadToStr: array[Boolean] of String = ('Output', 'Input');

var
  I, J: Integer;
  Buf: String;

begin
  ViewList.Items.Clear;
  for I := 0 to High(PortServ) do begin
    PortServ[I].DoCleanup;

    if PortServ[I].bRead then
      Buf := 'Input'
    else
      Buf := 'Output'
    ;
    J := PortServ[I].ServerSock.ListenPort;
    Buf := {IntToStr(I) + ': ' +} Buf + ' on port ' + IntToHex(J, 4) + 'h (';
    if (J and $FFF0) = $ED10 then
      Buf := Buf + 'read'
    else if (J and $FFF0) = $ED00 then
      Buf := Buf + 'write'
    else
      Buf := Buf + '???'
    ;
    Buf := Buf + ' index ' + IntToStr(J and $F) + ')';
    ViewList.Items.Add(Buf);

    if PortServ[I].ServerSock.ListenThread = nil then
      Buf := '  ENET NOT LISTENING'
    else
      Buf := '  ENet listening'
    ;
    J := Length(PortServ[I].ServerSock.ServerClients);
    if J <> 0 then begin
      Buf := Buf + ' (' + IntToStr(J) + ' connection' + StringOfChar('s', Ord(J <> 1)) + ')';
    end;
    ViewList.Items.Add(Buf);

    if not PortServ[I].bWorking then
      ViewList.Items.Add('  DIO NOT WORKING (' + IntToStr(PortServ[I].Status) + ')')
    else begin
      ViewList.Items.Add('  DIO working');
      J := PortServ[I].LastHz;
      if J = 0 then
        ViewList.Items.Add('  External/Off')
      else
        ViewList.Items.Add('  ' + IntToStr(J) + ' Hz')
      ;

      J := PortServ[I].PacketBits;
      ViewList.Items.Add('  Packet Bits: ' + IntToStr(J) + 'b (' + IntToStr((J + 7) div 8) + 'B)');
      ViewList.Items.Add('  Packets Queued: ' + IntToStr(PortServ[I].PacketFIFO.PeekDepth));

      if PortServ[I].bRead then begin
      J := PortServ[I].iPacketState;
        case J of
          0: ViewList.Items.Add('  Packet State: Not Yet');
          1: ViewList.Items.Add('  Packet State: Receiving Packet');
          2: ViewList.Items.Add('  Packet State: Between Packets');
        end;
      end;
    end;
    ViewList.Items.Add('');
  end;
  Buf := '';
end;

end.
